unit OwnedList;

{
    Copyright (C) 2011 Ondrej Pokorny
      http://www.kluug.at
}

{$I OWideSupp.inc}

interface

uses
  Classes;

type
  TOwnedListObject = class;
  TOwnedList = class;

  TOwnedListNotify = procedure (Sender: TOwnedList; Item: TOwnedListObject; Action: TListNotification) of Object;

  TOwnedList = class(TList)
  private
    fOwnerObject: TObject;
    fOnNotify: TOwnedListNotify;

    function GetI(Index: Integer): TOwnedListObject;
    procedure PutI(Index: Integer; const Value: TOwnedListObject);
  protected
    procedure Notify(Ptr: Pointer; Action: TListNotification); override;

    procedure DoCreate; virtual;
    procedure DoDestroy; virtual;
  public
    constructor Create;
    destructor Destroy; override;
  public
    procedure Assign(FromList: TOwnedList); virtual;
    procedure AssignCreate(FromList: TOwnedList);
    procedure AddCreate(FromList: TOwnedList); virtual;
    procedure InsertCreate(FromList: TOwnedList; ToIndex: Integer); virtual;
  public
    function Add(Item: Pointer): Integer; reintroduce; virtual;
    procedure Delete(Index: Integer); reintroduce; virtual;
    procedure Exchange(Index1, Index2: Integer); reintroduce; virtual;
    function Extract(Item: Pointer): Pointer; reintroduce; virtual;
    function First: TOwnedListObject; reintroduce;
    function IndexOf(Item: Pointer): Integer; reintroduce; virtual;
    procedure Insert(Index: Integer; Item: Pointer); reintroduce; virtual;
    function Last: TOwnedListObject; reintroduce;
    procedure Move(CurIndex, NewIndex: Integer); reintroduce; virtual;
    function Remove(Item: Pointer): Integer; reintroduce; virtual;

    property Items[Index: Integer]: TOwnedListObject read GetI write PutI; default;

    property OwnerObject: TObject read fOwnerObject write fOwnerObject;//f.e. if this list ist owned by another object

    property OnNotify: TOwnedListNotify read fOnNotify write fOnNotify;
  end;

  TOwnedListObject = class(TPersistent)
  private
    fOwnerList: TOwnedList;
    function GetIndexInList: Integer;
    procedure SetIndexInList(const Value: Integer);
    function GetOwnerObject: TObject;
  protected
    procedure AssignTo(Dest: TPersistent); override;

    procedure DoSetIndexInList(const OldIndex, NewIndex: Integer); virtual;

    procedure DoCreate; virtual;
    procedure DoDestroy; virtual;
  public
    constructor Create;
    constructor CreateAdd(AddTo: TOwnedList); overload;
    constructor CreateAdd(AddTo: TOwnedList; out Position: Integer); overload;
    constructor CreateInsert(AddTo: TOwnedList; Position: Integer);
    destructor Destroy; override;
  public
    property OwnerList: TOwnedList read fOwnerList;
    property OwnerObject: TObject read GetOwnerObject;

    property IndexInList: Integer read GetIndexInList write SetIndexInList;
  end;

  TOwnedListObjectClass = class of TOwnedListObject;

implementation

uses
  {$IFDEF O_DELPHI_XE2}
  Vcl.Dialogs
  {$ELSE}
  Dialogs
  {$ENDIF}
  ;


{ TOwnedList }

function TOwnedList.Add(Item: Pointer): Integer;
begin
  Result := inherited Add(Item);
end;

procedure TOwnedList.AddCreate(FromList: TOwnedList);
var I: Integer;
  Old, New: TObject;
begin
  for I := 0 to FromList.Count-1 do begin
    Old := TObject(FromList.Items[I]);

    if Old is TOwnedListObject then begin
      New := (TOwnedListObjectClass(Old.ClassType)).CreateAdd(Self);

      TOwnedListObject(New).Assign(TOwnedListObject(Old));
      //TOwnedListObject(Old).AssignTo(TOwnedListObject(New));
    end;
  end;
end;

procedure TOwnedList.Assign(FromList: TOwnedList);
begin
  AssignCreate(FromList);
end;

procedure TOwnedList.AssignCreate(FromList: TOwnedList);
begin
  Clear;
  AddCreate(FromList);
end;

constructor TOwnedList.Create;
begin
  inherited;

  DoCreate;
end;

procedure TOwnedList.Delete(Index: Integer);
begin
  inherited Delete(Index);
end;

destructor TOwnedList.Destroy;
begin
  DoDestroy;
  
  inherited;
end;

procedure TOwnedList.DoCreate;
begin
  fOwnerObject := nil;
  fOnNotify := nil;
end;

procedure TOwnedList.DoDestroy;
begin

end;

procedure TOwnedList.Exchange(Index1, Index2: Integer);
begin
  inherited Exchange(Index1, Index2);
end;

function TOwnedList.Extract(Item: Pointer): Pointer;
begin
  Result := inherited Extract(Item);
end;

function TOwnedList.First: TOwnedListObject;
begin
  Result := TOwnedListObject(inherited First);
end;

function TOwnedList.GetI(Index: Integer): TOwnedListObject;
begin
  Result := TOwnedListObject(inherited Items[Index]);
end;

function TOwnedList.IndexOf(Item: Pointer): Integer;
begin
  Result := inherited IndexOf(Item);
end;

procedure TOwnedList.Insert(Index: Integer; Item: Pointer);
begin
  inherited Insert(Index, Item);
end;

procedure TOwnedList.InsertCreate(FromList: TOwnedList; ToIndex: Integer);
var I: Integer;
  Old, New: TObject;
begin
  for I := 0 to FromList.Count-1 do begin
    Old := TObject(FromList.Items[I]);

    if Old is TOwnedListObject then begin
      New := (TOwnedListObjectClass(Old.ClassType)).CreateInsert(Self, ToIndex+I);

      TOwnedListObject(New).Assign(TOwnedListObject(Old));
      //TOwnedListObject(Old).AssignTo(TOwnedListObject(New));
    end;
  end;
end;

function TOwnedList.Last: TOwnedListObject;
begin
  Result := TOwnedListObject(inherited Last);
end;

procedure TOwnedList.Move(CurIndex, NewIndex: Integer);
begin
  inherited Move(CurIndex, NewIndex);
end;

procedure TOwnedList.Notify(Ptr: Pointer; Action: TListNotification);
begin
  if Assigned(fOnNotify) then
    fOnNotify(Self, TOwnedListObject(Ptr), Action);

  if Action = lnDeleted then
    TObject(Ptr).Free;

  inherited Notify(Ptr, Action);
end;

procedure TOwnedList.PutI(Index: Integer; const Value: TOwnedListObject);
begin
  inherited Items[Index] := Value;
end;

function TOwnedList.Remove(Item: Pointer): Integer;
begin
  Result := inherited Remove(Item);
end;

{ TOwnedListObject }

procedure TOwnedListObject.AssignTo(Dest: TPersistent);
begin
  //do nothing
end;

constructor TOwnedListObject.Create;
begin
  fOwnerList := nil;

  DoCreate;
end;

constructor TOwnedListObject.CreateAdd(AddTo: TOwnedList;
  out Position: Integer);
begin
  fOwnerList := AddTo;

  Position := AddTo.Add(Self);

  DoCreate;
end;

constructor TOwnedListObject.CreateAdd(AddTo: TOwnedList);
var Pos: Integer;
begin
  CreateAdd(AddTo, Pos);
end;

destructor TOwnedListObject.Destroy;
begin
  DoDestroy;
  
  inherited;
end;

procedure TOwnedListObject.DoCreate;
begin
  //nothing
end;

procedure TOwnedListObject.DoDestroy;
begin
  //nothing in this class
end;

procedure TOwnedListObject.DoSetIndexInList(const OldIndex, NewIndex: Integer);
begin
  OwnerList.Move(OldIndex, NewIndex);
end;

function TOwnedListObject.GetIndexInList: Integer;
begin
  if fOwnerList <> nil then
    Result := fOwnerList.IndexOf(Self) else
    Result := -1;
end;

function TOwnedListObject.GetOwnerObject: TObject;
begin
  if Assigned(fOwnerList) then
    Result := fOwnerList.OwnerObject
  else
    Result := nil;
end;

procedure TOwnedListObject.SetIndexInList(const Value: Integer);
var
  CurIndexInList: Integer;
begin
  CurIndexInList := GetIndexInList;
  if (CurIndexInList >= 0) and (CurIndexInList <> Value) then
  begin
    DoSetIndexInList(CurIndexInList, Value);
  end;
end;

constructor TOwnedListObject.CreateInsert(AddTo: TOwnedList; Position: Integer);
begin
  fOwnerList := AddTo;
  if Position < 0 then
    Position := 0;
  if Position > AddTo.Count then
    Position := AddTo.Count;

  DoCreate;

  AddTo.Insert(Position, Self);
end;

end.
